from .models import TotalStats
from string import Template
import re

TPL_VAR_PATTERN = r'\$\{([a-zA-Z0-9_]+(?:\_[a-zA-Z0-9_]+)*)\}'

def get_val(d, k):
    keys = k.split('_')
    for key in keys:
        if key.isdigit():
            if key == keys[-1]:
                try:
                    return list(d.keys())[int(key)]
                except:
                    return "-"
            else:
                try:
                    d = d[list(d.keys())[int(key)]]
                except:
                    return "-"
        elif isinstance(d, dict) and key in d:
            d = d[key]
        else:
            return None
    if key == "percent":
        return "{d}%".format(d=d)
    return d

# Create HTML reports with stats
class Html:

    tpl: str = ""
    stats: TotalStats

    def __init__(self, tpl_file: str, stats: TotalStats):
        self.tpl_file = tpl_file
        self.stats = stats.dict()

        with open(tpl_file, 'r') as tpl_html:
            self.tpl = tpl_html.read()

    def build(self):
        matches = re.findall(TPL_VAR_PATTERN, self.tpl)
        replacements = {}
        for match in matches:
            value = get_val(self.stats, match)
            if isinstance(value, float):
                value = round(value, 2)
            replacements[match] = value
        
        template = Template(self.tpl)
        return template.substitute(replacements)

    def dump(self):
        print(self.build())


from pyproj import Geod
from shapely.geometry import MultiLineString, Polygon, shape

# Utilities
class GeoUtils:
    def __init__(self, projected: bool, proj: str = "WGS84"):
        self.projected = projected
        self.proj = proj

    # Calculate way's length in km
    def way_length(self, way: object):
        geo: dict = way["geometry"]
        line: MultiLineString = shape(geo)
        if self.projected:
            length = line.length / 1000
        else:
            geod = Geod(ellps=self.proj)
            length = geod.geometry_length(line) / 1000
        return length 

    # Calculate way's area in km2
    def way_area(self, way: object):
        geo: dict = way["geometry"]
        polygon: Polygon = shape(geo)
        if self.projected:
            area = polygon.area / 1000000
        else:
            geod = Geod(ellps=self.proj)
            area = abs(geod.geometry_area_perimeter(polygon)[0]) / 1000000
        return area 



import os
import json
from urllib.request import Request, urlopen
from .geoutils import GeoUtils
from .models import *

# Stats generator
class Stats:

    def __init__(self, config: Config = None):
        if config is None:
            config = Config()
        self.config = config
        self.cache = CalcCache()
        self.results = TotalStats()
        self.geo_utils = GeoUtils(
            projected=self.config.projected
        )

    # Get property from tag
    def getProperty(self, path: str, obj: dict):
        keys = path.split('.')
        value = obj
        
        for key in keys:
            value = value.get(key)
            if value is None:
                return None
        
        return value

    # Clean calculation cache
    def clean_cache(self):
        self.cache = self.cache = CalcCache()

    # Process a line of a file
    def process_file_line(self, line: str):
        if line[-2:-1] == ",":
            json_string = line[:-2]
        else:
            json_string = line
        json_object = json.loads(json_string)
        self.get_object_stats(json_object)

    # Count value by key
    def count_value_bykey(self, json_object: object, key: str, value: str):
        if value and key in self.config.value_keys:
            self.results.bykeyvalue(key, value).sum()

    # Calculate total area in km2
    def calculate_area(self, json_object: object): 
        if json_object["geometry"] and (json_object["geometry"]["type"] == "Polygon" \
            or json_object["geometry"]["type"] == "MultiPolygon"):
                self.results.sum_area(self.geo_utils.way_area(json_object))

    # Calculate length in km by key
    def calculate_length_bykey(self, json_object: object, key: str):
        if key in self.config.keys:
            if not self.cache.way_length:
                self.cache.way_length = self.geo_utils.way_length(json_object)
            self.results.bykey(key).sum_length(self.cache.way_length)

    # Calculate length in km by key value
    def calculate_length_bykeyval(self, json_object: object, key: str, val: str):
        if val and key in self.config.value_keys:
            self.results.bykeyvalue(key, val).sum_length(self.geo_utils.way_length(json_object))

    # Calculate area in km2 by key
    def calculate_area_bykey(self, json_object: object, key: str):
        if key in self.config.keys:
            if not self.cache.way_area:
                self.cache.way_area = self.geo_utils.way_area(json_object)
            self.results.bykey(key).sum_area(self.cache.way_area)

    # Calculate area in km2 by key value
    def calculate_area_bykeyval(self, json_object: object, key: str, val: str):
        if val and key in self.config.value_keys:
            self.results.bykeyvalue(key, val).sum_area(self.geo_utils.way_area(json_object))

    # Calculate total length in km
    def calculate_length(self, json_object: object):
        if (json_object["geometry"]["type"] == "LineString" \
             or json_object["geometry"]["type"] == "MultiLineString"):
                self.results.sum_length(self.geo_utils.way_length(json_object))

    # Total count by key
    def total_keys(self, key: str):
        self.results.bykey(key).sum()

    # Get stats for a Feature object
    def get_object_stats(self, json_object: object):
        for prop in self.getProperty(self.config.properties_prop, json_object).items():
            key = prop[0]
            value = prop[1]

            if value:
                self.total_keys(key)

            if self.config.keys and prop[0] in self.config.keys:
                if self.config.length:
                    self.calculate_length_bykey(json_object,key)
                if self.config.area:
                    self.calculate_area_bykey(json_object, key)

            if self.config.value_keys and key in self.config.value_keys:
                self.count_value_bykey(json_object, key, value)
                if self.config.length:
                    self.calculate_length_bykeyval(json_object, key, value)
                if self.config.area:
                    self.calculate_area_bykeyval(json_object, key, value)

            if key == "name" and not "name" in self.results.languages:
                self.results.languages.append("name")

            if len(key) == 7 and "name:" in key and key.index("name:") == 0:
                if not key in self.results.languages:
                    self.results.languages.append(key)

        if self.config.length:
            self.calculate_length(json_object)

        if self.config.area:
            self.calculate_area(json_object)

        self.results.sum()

        self.clean_cache()

    # Process a GeoJSON object
    def process_geojson(self, geojson_object: object):
        features_total = len(geojson_object["features"])
        for feature in geojson_object["features"]:
            self.get_object_stats(feature)
            if self.config.verbose:
                percent = round((self.results.count * 100) / features_total, 2)
                print("Processed: {0}% ({1})".format(percent, self.results.count),\
                        end='\r', flush=True)

    # Process a GeoJSON file line by line
    def process_file_stream(self, filename: str):
        file = open(filename)
        bytes_total = os.stat(filename).st_size
        bytes_processed = 0
        for line in file:
            bytes_processed += len(line)
            if line.startswith('{ "type": "Feature"'):
                self.process_file_line(line)
                percent = round((bytes_processed * 100) / bytes_total, 2)
                if self.config.verbose:
                    print("Processed: {0}% ({1})".format(percent, self.results.count),\
                          end='\r', flush=True)

    # Process a GeoJSON file
    def process_file(self, filename: str):
        if self.config.verbose:
            print("Opening file ...\n")
        with open(filename, 'r') as json_data:
            json_data = json.load(json_data)
            self.process_geojson(json_data)

    # Process an URL
    def process_url(self, url: str):
        if self.config.verbose:
            print("Downloading file from URL ...\n")

        req = Request(
            url=url,
            headers={'User-Agent': 'Mozilla/5.0'})

        json_data = urlopen(req).read().decode('utf-8')
        self.process_geojson(json.loads(json_data))

    # Returns a JSON string with the results
    def json(self):
        return json.dumps(self.dict())

    # Dumps results
    def dump(self):
        print(self.json())

    def dict(self):
        return self.results.to_dict(clean=self.config.clean)



#!/usr/bin/python3

import argparse
import resource
import os

from .stats import Config, Stats
from .html import Html

# Entrypoint for command line
def main():
    args = argparse.ArgumentParser()
    args.add_argument("--file", "-f", help="GeoJSON file to analyze", type=str, default=None)
    args.add_argument("--url", "-u", help="URL of GeoJSON file to analyze", type=str, default=None)
    args.add_argument("--verbose", "-v", help="Verbose", default=False, action='store_true')
    args.add_argument("--stream", help="Stream a file (use less memory)", default=False, action='store_true')
    args.add_argument("--keys", help="Keys for calculating length or area", type=str, default = None)
    args.add_argument("--value-keys", help="Keys for counting values", type=str, default = None)
    args.add_argument("--length", help="Calculate total length of all linestrings", \
                    default=False, action='store_true')
    args.add_argument("--area", help="Calculate total area of all polygons", \
                    default=False, action='store_true')
    args.add_argument("--projected", help="Use projected coordinated in meters", \
                    default=False, action='store_true')
    args.add_argument("--proj", help="Data projection system", type=str, default = "WGS84")
    args.add_argument("--no-clean", help="Keep zero and empty results", default=False, action='store_true')
    args.add_argument("--properties-prop", help="Properties to analyze (ex: properties or properties.tags)", type=str, default="properties")
    args.add_argument("--html", help="HTML template", type=str, default=None)
    args = args.parse_args()

    config = Config(
        verbose=args.verbose,
        keys = args.keys.split(",") if args.keys else [],
        length = args.length,
        area = args.area,
        value_keys = args.value_keys,
        projected = args.projected,
        clean = not args.no_clean,
        proj = args.proj,
        properties_prop = args.properties_prop
    )
    stats = Stats(config)

    if args.url or args.file:

        if args.url:
            stats.process_url(args.url)

        elif args.file:
            if args.verbose:
                print("\nFile size is {0} MB\n".format(round(os.stat(args.file).st_size / (1024 * 1024), 2)))
            if args.stream:
                stats.process_file_stream(args.file)
            else:
                stats.process_file(args.file)

        if args.html:
            html = Html(args.html, stats)
            html.dump()
        else:
            stats.dump()

        if args.verbose:
            print('\nPeak Memory Usage =', resource.getrusage(resource.RUSAGE_SELF).ru_maxrss)
            print('User Mode Time =', resource.getrusage(resource.RUSAGE_SELF).ru_utime)
            print('System Mode Time =', resource.getrusage(resource.RUSAGE_SELF).ru_stime)

if __name__ == "__main__":
    main()



from dataclasses import dataclass, field

@dataclass
class Config:
    # Be verbose
    verbose: bool = False
    # Keys for measure length or area
    keys: list = field(default_factory=lambda: [])
    # Enable/disable length calculation
    length: bool = False
    # Enable/disable area calculation
    area: bool = False
    # Keys for collecting value stats
    value_keys: list = field(default_factory=lambda: [])
    # Use projected coordinates
    projected: bool = False
    # Remove zero and empty results
    clean: bool = False
    # Coordinates system
    proj: str = "WGS84"
    # Properties to analyze (ex: "properties" or "properties.tags")
    properties_prop: str = "properties"


# Base stats class
@dataclass
class BaseStats:
    count: int = 0
    length: float = 0
    area: float = 0

    def sum(self, key = None):
        self.count += 1

    def sum_length(self, length: float = 0):
        self.length += length

    def sum_area(self, area: float = 0):
        self.area += area

    def to_dict(self, clean = False):
        res = {
            "count": self.count,
            "area": self.area,
            "length": self.length
        }
        if clean:
            if res["count"] == 0:
                del res["count"]
            if res["area"] == 0:
                del res["area"]
            if res["length"] == 0:
                del res["length"]
        return res

# Stats by value
@dataclass
class ValueStats(BaseStats):
    pass

# Stats by key
@dataclass
class KeyStats(BaseStats):
    value: dict = field(default_factory=lambda: {})

    def to_dict(self, clean = False, total = 1):
        res = {
            "count": self.count,
            "percent": round(self.count * 100 / total, 2),
            "area": self.area,
            "length": self.length,
            "value": {k:v.to_dict(clean) for k, v in self.value.items()}
        }
        if clean:
            if res["count"] == 0:
                del res["count"]
            if res["area"] == 0:
                del res["area"]
            if res["length"] == 0:
                del res["length"]
            if res["value"] == {}:
                del res["value"]
        return res

# Total stats
@dataclass
class TotalStats(BaseStats):
    key: dict = field(default_factory=lambda: {})
    languages: list = field(default_factory=lambda: [])

    def bykey(self, key: str):
        if key not in self.key:
            self.key[key] = KeyStats()
        return self.key[key]

    def bykeyvalue(self, key: str, value: str):
        key_object = self.bykey(key)
        if not value in key_object.value:
            key_object.value[value] = ValueStats()
        return self.key[key].value[value]

    def to_dict(self, clean = False):
        res = {
            "count": self.count,
            "area": self.area,
            "length": self.length,
            "languages": {
                "list": self.languages,
                "count": len(self.languages),
            },
            "key": {k:v.to_dict(clean, self.count) for k, v in self.key.items()}
        }
        if clean:
            if res["count"] == 0:
                del res["count"]
            if res["area"] == 0:
                del res["area"]
            if res["length"] == 0:
                del res["length"]
        res["key"] = dict(sorted(res["key"].items(), key=lambda x: x[1]["count"], reverse=True))
        return res


# Cache for storing calculations
@dataclass
class CalcCache:
    way_length: float = None
    way_area: float = None